from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import zodiax


def sample_function(pytree):
    return pytree.param * pytree.b.param


# Define paths
path1 = 'param'
path2 = 'b.param'


def test_filter_grad(Base_instance):
    pytree = Base_instance

    args = [path1]
    filter_spec = zodiax.tree.get_args(pytree, args)

    _ = zodiax.equinox.filter_grad(filter_spec)(sample_function)(pytree)
    _ = zodiax.equinox.filter_grad(args)(sample_function)(pytree)
    

def test_filter_value_and_grad(Base_instance):
    pytree = Base_instance

    args = [path1]
    filter_spec = zodiax.tree.get_args(pytree, args)

    _ = zodiax.equinox.filter_value_and_grad(filter_spec)(sample_function)(pytree)
    _ = zodiax.equinox.filter_value_and_grad(args)(sample_function)(pytree)


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
from optax import adam, GradientTransformation, MultiTransformState
import zodiax


# Define paths
path1 = 'param'
path2 = 'b.param'


def test_get_optimiser(Base_instance):
    """
    tests the get_optimiser method
    """
    # Define parameters and construct base
    base = Base_instance

    # Define paths & groups
    optimisers = [adam(0), adam(1)] # These are actually arbitrary

    # Test paths
    optim, opt_state = zodiax.optimisation.get_optimiser(base, [path1, path2], optimisers)
    assert isinstance(optim, GradientTransformation)
    assert isinstance(opt_state, MultiTransformState)


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import zodiax


# Define paths
path1 = 'param'
path2 = 'b.param'


def test_get_args(Base_instance):
    pytree = Base_instance

    filter_spec = zodiax.tree.get_args(pytree, [path1])

    assert filter_spec.get(path1) == True
    assert filter_spec.get(path2) == False


name = "zodiax"
__version__ = "0.3.3"

# Import as modules
from . import base
from . import tree
from . import optimisation
from . import experimental
from . import equinox

# Import core functions from modules
from .base import *
from .tree import *
from .optimisation import *
from .equinox import *

# Add to __all__
__all__ = base.__all__ + tree.__all__ + optimisation.__all__ + \
    experimental.__all__ + equinox.__all__



from __future__ import annotations
import zodiax
import jax.numpy as np
from equinox import tree_at, Module
from typing import Union, Any, Callable, List
from jaxtyping import Array, PyTree


__all__ = ["Base"]

PathLike = Union[List[str], str]


def _get_leaf(pytree : PyTree, path : PathLike) -> Any:
    """
    A hidden class desinged to recurse down a pytree following the path,
    returning the leaf at the end of the path.

    Base case: len(path) == 1
        In this case the leaf referred to by the single path entry is
        returned (and hence recursively sent up to the initial call).

    Recursive case: len(path) > 1
        In this case the function takes the PyTree like object referred to
        by the first entry in path, and recursively calls this function
        with this new pytree object and the path without the first entry.

    Parameters
    ----------
    pytree : PyTree
        The pytee object to recurse though.
    path : PathLike
        The path to recurse down.

    Returns
    -------
    leaf : Any
        The leaf object specified at the end of the path object.
    """
    key = path[0]
    if hasattr(pytree, key):
        pytree = getattr(pytree, key)
    elif isinstance(pytree, dict):
        pytree = pytree[key]
    elif isinstance(pytree, (list, tuple)):
        pytree = pytree[int(key)]
    else:
        raise ValueError(
            "key: {} not found in object: {}".format(key,type(pytree)))

    # Return param if at the end of path, else recurse
    return pytree if len(path) == 1 else _get_leaf(pytree, path[1:])


def _get_leaves(pytree : PyTree, paths : list) -> list:
    """
    Returns a list of leaves specified by the paths.

    Parameters
    ----------
    pytree : PyTree
        The pytee object to recurse though.
    paths : list
        A list/tuple of nested paths. Note path objects can only be
        nested a single time.

    Returns
    -------
    leaves : list
        The list of leaf objects specified by the paths object
    """
    return [_get_leaf(pytree, path) for path in paths]


def _unwrap(paths : PathLike, values_in : list = None) -> list:
    """
    Unwraps the provided paths in to the correct list-based format for the
    _get_leaves and _get_leaf methods, returning a single dimensional list
    of input paths.

    Parameters
    ----------
    paths : PathLike
        A list/tuple of nested paths to unwrap.
    values_in : list = None
        The list of values to be unwrapped.

    Returns
    -------
    paths, values : list, list
        The list of unwrapped paths or paths and values.
    """
    # Inititalise empty lists
    paths_out, values_out = [], []

    # If values are provided, apply transformation to both
    if values_in is not None:
        # Make sure values is list
        values = values_in if isinstance(values_in, list) else [values_in]

        # Repeat values to match length of paths
        if len(values) == 1:
            values = values * len(paths)
        
        # Ensure correct length
        if len(values) != len(paths):
            raise ValueError(
                "The number of values must match the number of paths.")

        # Iterate over paths and values
        for path, value in zip(paths, values):

            # Recurse and add in the case of list inputs
            if isinstance(path, list):
                new_paths, new_values = _unwrap(path, value)
                paths_out  += new_paths
                values_out += new_values

            # PathLike must already be absolute
            else:
                paths_out.append(path)
                values_out.append(value)
        return paths_out, values_out

    # Just paths provided
    else:
        # Iterate over paths
        for path in paths:

            # Recurse and add in the case of list inputs
            if isinstance(path, list):
                new_paths = _unwrap(path)
                paths_out += new_paths

            # PathLike must already be absolute
            else:
                paths_out.append(path)
        return paths_out


def _format(paths : PathLike, values : list = None) -> list:
    """
    Formats the provided paths in to the correct list-based format for the
    _get_leaves and _get_leaf methods, returning a single dimensional list
    of input paths.

    Parameters
    ----------
    paths : PathLike
        A list/tuple of nested paths to unwrap.
    values : list = None
        The list of values to be unwrapped.

    Returns
    -------
    paths, values : list, list
        The list of unwrapped paths or paths and values.
    """
    # Nested/multiple inputs
    if isinstance(paths, list):

        # If there is nesting, ensure correct dis
        if len(paths) > 1 and values is not None \
            and True in [isinstance(p, list) for p in paths]:
            assert isinstance(values, list) and len(values) == len(paths), \
            ("If a list of paths is provided, the list of values must be "
                "of equal length.")

        # Its a list - iterate and unbind all the keys
        if values is not None:
            flat_paths, new_values = _unwrap(paths, values)
        else:
            flat_paths = _unwrap(paths)
        
        # Turn into seperate strings
        new_paths = [path.split('.') if '.' in path else [path] \
                        for path in flat_paths]

    # Un-nested/singular input
    else:
        # Turn into seperate strings
        new_paths = [paths.split('.') if '.' in paths else [paths]]
        new_values = [values]

    # Return
    return new_paths if values is None else (new_paths, new_values)


###############
### Classes ###
###############
class Base(Module):
    """
    Extend the Equninox.Module class to give a user-friendly 'path based' API
    for working with pytrees by adding a series of methods used to interface
    with the leaves of the pytree using paths.
    """
    def get(self : PyTree, paths : PathLike) -> Any:
        """
        Get the leaf specified by path.

        Parameters
        ----------
        paths : PathLike
            A list/tuple of nested paths to unwrap.

        Returns
        -------
        leaf, leaves : Any, list
            The leaf or list of leaves specified by paths.
        """
        new_paths = _format(paths)
        values = _get_leaves(self, new_paths)
        return values[0] if len(new_paths) == 1 else values


    def set(self   : PyTree,
            paths  : PathLike,
            values : Union[List[Any], Any]) -> PyTree:
        """
        Set the leaves specified by paths with values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to set at the leaves specified by paths.

        Returns
        -------
        pytree : PyTree
            The pytree with leaves specified by paths updated with values.
        """
        # Allow None inputs
        if values is None:
            values = [None]
            if isinstance(paths, str):
                paths = [paths]
        new_paths, new_values = _format(paths, values)

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def add(self   : PyTree,
            paths  : PathLike,
            values : Union[List[Any], Any]) -> PyTree:
        """
        Add to the the leaves specified by paths with values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to add to the leaves specified by paths.

        Returns
        -------
        pytree : PyTree
            The pytree with values added to leaves specified by paths.
        """
        new_paths, new_values = _format(paths, values)
        new_values = [leaf + value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def multiply(self   : PyTree,
                 paths  : PathLike,
                 values : Union[List[Any], Any]) -> PyTree:
        """
        Multiplies the the leaves specified by paths with values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to multiply the leaves specified by paths.

        Returns
        -------
        pytree : PyTree
            The pytree with values multiplied by leaves specified by paths.
        """
        new_paths, new_values = _format(paths, values)
        new_values = [leaf * value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def divide(self   : PyTree,
               paths  : PathLike,
               values : Union[List[Any], Any]) -> PyTree:
        """
        Divides the the leaves specified by paths with values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to divide the leaves specified by paths.

        Returns
        -------
        pytree : PyTree
            The pytree with values divided by leaves specified by paths.
        """
        new_paths, new_values = _format(paths, values)
        new_values = [leaf / value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def power(self   : PyTree,
              paths  : PathLike,
              values : Union[List[Any], Any]) -> PyTree:
        """
        Raises th leaves specified by paths to the power of values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to take the leaves specified by paths to the
            power of.

        Returns
        -------
        pytree : PyTree
            The pytree with the leaves specified by paths raised to the power
            of values.
        """
        new_paths, new_values = _format(paths, values)
        new_values = [leaf ** value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def min(self   : PyTree,
            paths  : PathLike,
            values : Union[List[Any], Any]) -> PyTree:
        """
        Updates the leaves specified by paths with the minimum value of the
        leaves and values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to take the minimum of and the leaf.

        Returns
        -------
        pytree : PyTree
            The pytree with the leaves specified by paths updated with the
            minimum value of the leaf and values.
        """
        new_paths, new_values = _format(paths, values)
        new_values = [np.minimum(leaf, value) for value, leaf in \
                    zip(new_values, _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def max(self   : PyTree,
            paths  : PathLike,
            values : Union[List[Any], Any]) -> PyTree:
        """
        Updates the leaves specified by paths with the maximum value of the
        leaves and values.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to take the maximum of and the leaf.

        Returns
        -------
        pytree : PyTree
            The pytree with the leaves specified by paths updated with the
            maximum value of the leaf and values.
        """
        new_paths, new_values = _format(paths, values)
        new_values = [np.maximum(leaf, value) for value, leaf in \
                    zip(new_values, _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def apply(self  : PyTree,
              paths : PathLike,
              fns   : Union[List[Callable], Callable]) -> PyTree:
        """
        Applies the functions within fns the leaves specified by paths.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        fns : Union[List[Callable], Callable]
            The list of functions to apply to the leaves.

        Returns
        -------
        pytree : PyTree
            The pytree with fns applied to the leaves specified by paths.
        """
        new_paths, new_fns = _format(paths, fns)
        new_values = [fn(leaf) for fn, leaf in zip(new_fns, \
                                    _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def apply_args(self  : PyTree,
                   paths : PathLike,
                   fns   : Union[List[Callable], Callable],
                   args  : Union[List[Any], Any]) -> PyTree:
        """
        Applies the functions within fns the leaves specified by paths, while
        also passing in args to the function.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        fns : Union[List[Callable], Callable]
            The list of functions to apply to the leaves.
        args : Union[List[Any], Any]
            The tupe or list of tuples of extra arguments to pass into fns.

        Returns
        -------
        pytree : PyTree
            The pytree with fns applied to the leaves specified by paths with
            the extra args passed in.
        """
        new_paths, new_fns = _format(paths, fns)
        new_paths, new_args = _format(paths, args)
        new_values = [fn(leaf, *args) for fn, args, leaf in zip(new_fns, \
                            new_args, _get_leaves(self, new_paths))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_paths)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def set_and_call(self    : PyTree,
                     paths   : PathLike,
                     values  : Union[List[Any], Any],
                     call_fn : str,
                     **kwargs) -> Any:
        """
        Updates the leaves speficied by paths with values, and then calls the
        function specified by the string call_fn, returning whatever is
        returnd by the call_fn. Any extra positional arguments or key-word
        arguments are passed through to the modelling function.

        This function is desigend to be used in conjunction with numpyro.
        Please go through the 'PyTree interface' tutorial to see how this
        is used.

        Parameters
        ----------
        paths : PathLike
            A path or list of paths or list of nested paths.
        values : Union[List[Any], Any]
            The list of values to set at the leaves specified by paths.
        call_fn : str
            A string specifying which model function to call.

        Returns
        -------
            : Any
            Whatever object is returned by call_fn.
        """
        return getattr(self.set(paths, values), call_fn)(**kwargs)


    def apply_and_call(self     : PyTree,
                       paths    : PathLike,
                       fns      : Union[List[Callable], Callable],
                       call_fn  : str,
                       **kwargs) -> object:
        """
        Applies the functions specified by fns to the leaves speficied by
        paths, and then calls the function specified by the string call_fn,
        returning whatever is returnd by the call_fn. Any extra positional
        arguments or keyword arguments are passed through to the modelling
        function.

        Parameters
        ----------
        call_fn : str
            A string specifying which model function to call.
        paths : PathLike
            A path or list of paths or list of nested paths.
        fns : Union[List[Callable], Callable]
            The list of functions to apply to the leaves.

        Returns
        -------
            : Any
            Whatever object is returned by call_fn.
        """
        return getattr(self.apply(paths, fns), call_fn)(**kwargs)


"""
Extra methods for possible future classes that recursively searches for 
parameters.
"""
#     Method 1, get first
#     def __getattr__(self, key):
#         """

#         """
#         # Found it, nice work
#         dict_like = self.__dict__
#         if key in dict_like.keys():
#             return dict_like[key]

#         # Expand and iterate though items
#         for value in dict_like.values():

#             # Dictionary, call the recursive method
#             if isinstance(value, dict):
#                 try:
#                     return _recurse_dict(value, key)
#                 except ValueError as e:
#                     pass

#             # dLux object, recurse
#             if isinstance(value, Base):
#                 try:
#                     return getattr(value, key)
#                 except ValueError as e:
#                     pass

#         # Not found, raise error
#         raise ValueError("'{}' object has no attribute '{}'"\
#                              .format(type(self), key))


#     def _recurse_dict(self, dict_like, key):
#         """

#         """
#         # Return item if it exists
#         if key in dict_like:
#             return dict_like[key]

#         # Iterate through values
#         for value in dict_like.values():

#             # Value is a dict, Recurse
#             if isinstance(value, dict):
#                 try:
#                     return self._recurse_dict(value, key)
#                 except ValueError as e:
#                     pass

#             # Value is a dLux object, recall the getattr method
#             if isinstance(value, Base):
#                 try:
#                     return getattr(value, key)
#                 except ValueError as e:
#                     pass

#         # Nothing found, raise Error
#         raise ValueError("'{}' not found.".format(key))


#     # Method 2, get all
#     def __getattr__(self, key):
#         """

#         """
#         return self._get_all(key, [])


#     def _get_all(self, key, values):

#         # Found it, nice work
#         dict_like = self.__dict__
#         if key in dict_like.keys():
#             values.append(dict_like[key])

#         # Expand and iterate though items
#         for value in dict_like.values():

#             # Dictionary, call the recursive method
#             if isinstance(value, dict):
#                 # values.append(self._recurse_dict(value, key, values))
#                 values = self._recurse_dict(value, key, values)

#             # dLux object, recurse
#             if isinstance(value, Base):
#                 # values.append(value._get_all(key, values))
#                 values = value._get_all(key, values)

#         return values


#     def _recurse_dict(self, dict_like, key, values):
#         """

#         """
#         # Return item if it exists
#         if key in dict_like:
#             values.append(dict_like[key])

#         # Iterate through values
#         for value in dict_like.values():

#             # Value is a dict, Recurse
#             if isinstance(value, dict):
#                 # values.append(self._recurse_dict(value, key, values))
#                 values = self._recurse_dict(value, key, values)

#             # Value is a dLux object, recall the getattr method
#             if isinstance(value, Base):
#                 # values.append(value._get_all(key, values))
#                 values = value._get_all(key, values)

#         return values


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import jax.numpy as np
import zodiax
import pytest


def test_unwrap():
    """
    Test the _unwrap method
    """

    # Test unwrapping
    wrapped_a = ['a', ['b', ['c', ['d']]]]
    wrapped_b = [[[['a'], 'b'], 'c'], 'd']
    wrapped_c = ['a', 'b', 'c', 'd']

    assert zodiax.base._unwrap(wrapped_a) == ['a', 'b', 'c', 'd']
    assert zodiax.base._unwrap(wrapped_b) == ['a', 'b', 'c', 'd']
    assert zodiax.base._unwrap(wrapped_c) == ['a', 'b', 'c', 'd']

    # Test with values
    wrapped_a = ['a', ['b', 'c', 'd']]
    wrapped_b = [['a', 'b', 'c'], 'd']
    wrapped_c = ['a', 'b', 'c', 'd']

    assert zodiax.base._unwrap(wrapped_a, [1, 2])[1]       == [1, 2, 2, 2]
    assert zodiax.base._unwrap(wrapped_b, [1, 2])[1]       == [1, 1, 1, 2]
    assert zodiax.base._unwrap(wrapped_c, [1, 2, 3, 4])[1] == [1, 2, 3, 4]


def test_format():
    """
    test the _format method
    """

    # Test formatting
    path_a = ['a.b', ['b.c', ['c.d', ['d.e']]]]
    path_b = [[[['a.b'], 'b.c'], 'c.d'], 'd.e']
    path_c = ['a.b', 'b.c', 'c.d', 'd.e']

    assert zodiax.base._format(path_a) == [['a', 'b'], ['b', 'c'], ['c', 'd'], ['d', 'e']]
    assert zodiax.base._format(path_b) == [['a', 'b'], ['b', 'c'], ['c', 'd'], ['d', 'e']]
    assert zodiax.base._format(path_c) == [['a', 'b'], ['b', 'c'], ['c', 'd'], ['d', 'e']]

    # Test with values
    path_a = ['a.b', ['b.c', 'c.d', 'd.e']]
    path_b = [['a.b', 'b.c', 'c.d'], 'd.e']
    path_c = ['a.b', 'b.c', 'c.d', 'd.e']

    assert zodiax.base._format(path_a, [1, 2])[1]       == [1, 2, 2, 2]
    assert zodiax.base._format(path_b, [1, 2])[1]       == [1, 1, 1, 2]
    assert zodiax.base._format(path_c, [1, 2, 3, 4])[1] == [1, 2, 3, 4]


# Define paths
path1 = 'param'
path2 = 'b.param'

@pytest.mark.usefixtures("Base_instance")
class TestBase():
    """
    Tests the Base class.
    """

    def test_get(self, Base_instance):
        """
        tests the get method
        """
        # Define parameters and construct base
        base = Base_instance

        assert base.get(path1) == 1.
        assert base.get(path2) == 2.


    def test_set(self, Base_instance):
        """
        tests the set method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test set
        new_base = base.set([path1, path2], [param1, param2])
        assert new_base.param   == param1
        assert new_base.b.param == param2


    def test_add(self, Base_instance):
        """
        tests the add method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test add
        new_base = base.add([path1, path2], [param1, param2])
        assert new_base.param   == base.get(path1) + param1
        assert new_base.b.param == base.get(path2) + param2


    def test_multiply(self, Base_instance):
        """
        tests the multiply method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test multiply
        new_base = base.multiply([path1, path2], [param1, param2])
        assert new_base.param   == base.get(path1) * param1
        assert new_base.b.param == base.get(path2) * param2


    def test_divide(self, Base_instance):
        """
        tests the divide method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test divide
        new_base = base.divide([path1, path2], [param1, param2])
        assert new_base.param   == base.get(path1) / param1
        assert new_base.b.param == base.get(path2) / param2


    def test_power(self, Base_instance):
        """
        tests the power method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test power
        new_base = base.power([path1, path2], [param1, param2])
        assert new_base.param   == base.get(path1)**param1
        assert new_base.b.param == base.get(path2)**param2


    def test_min(self, Base_instance):
        """
        tests the min method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test min
        new_base = base.min([path1, path2], [param1, param2])
        assert new_base.param   == np.minimum(base.get(path1), param1)
        assert new_base.b.param == np.minimum(base.get(path2), param2)


    def test_max(self, Base_instance):
        """
        tests the max method
        """
        # Define parameters and construct base
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test max
        new_base = base.max([path1, path2], [param1, param2])
        assert new_base.param   == np.maximum(base.get(path1), param1)
        assert new_base.b.param == np.maximum(base.get(path2), param2)


    def test_apply(self, Base_instance):
        """
        tests the  method
        """
        # Define parameters and construct base
        fn1 = lambda x: x + 5.
        fn2 = lambda x: x + 10.
        base = Base_instance

        # Test 
        new_base = base.apply([path1, path2], [fn1, fn2])
        assert new_base.param   == base.get(path1) + 5.
        assert new_base.b.param == base.get(path2) + 10.


    def test_apply_args(self, Base_instance):
        """
        tests the apply_args method
        """
        # Define parameters and construct base
        fn = lambda x, a: x + a
        param1 = 5.
        param2 = 10.
        base = Base_instance

        # Test apply_args
        new_base = base.apply_args([path1, path2], fn, [(param1,), (param2,)])
        assert new_base.param   == base.get(path1) + param1
        assert new_base.b.param == base.get(path2) + param2


    def test_set_and_call(self, Base_instance):
        """
        tests the set_and_call method
        """
        # Define parameters and construct base
        param1 = 2.
        param2 = 4.
        base = Base_instance

        # Define groups
        values = [param1, param2]

        # Test paths
        new_base = base.set_and_call([path1, path2], values, "model")
        assert new_base == param1**2 + param2**2

    def test_apply_and_calll(self, Base_instance):
        """
        tests the apply_and_calll method
        """
        # Define parameters and construct base
        base = Base_instance

        # Define paths & fns
        fn1 = lambda x: x * 2.
        fn2 = lambda x: x * 4.
        fns = [fn1, fn2]

        # Test paths
        new_base = base.apply_and_call([path1, path2], fns, "model")
        assert new_base == (base.get(path1)*2.)**2 + (base.get(path2)*4.)**2


from __future__ import annotations
from jax import config
import zodiax
from equinox import tree_equal
import os
config.update("jax_debug_nans", True)


def test_serialise_deserialise(Base_instance):
    pytree = Base_instance
    try:
        zodiax.experimental.serialisation.serialise('test_serialisation.zdx', pytree)
        # pytree_2 = zodiax.experimental.serialisation.deserialise('test_serialisation.zdx')
        
        os.remove('test_serialisation.zdx')
        # assert tree_equal(pytree, pytree_2)
    except Exception as e:
        try:
            os.remove('test_serialisation.zdx')
        except FileNotFoundError:
            pass
        raise e


import zodiax
import equinox
from functools import wraps
from jaxtyping import PyTree
from typing import Union, Callable
from equinox import partition, combine
from types import ModuleType


def _convert_to_filter(pytree : PyTree, 
                       params : Union[PyTree, list, str]) -> PyTree:
    """
    Converts the input params to a filter spec if not already.

    Parameters
    ----------
    pytree : PyTree
        The pytree to filter.
    params : Union[PyTree, list, str]
        The params to filter, or an existing filter.
    
    Returns
    -------
     : PyTree
        The filter spec.
    """
    if isinstance(params, zodiax.base.Base):
        return params
    else:
        return zodiax.tree.get_args(pytree, params)


def filter_grad(params : Union[PyTree, list, str], 
                *filter_args, **filter_kwargs) -> Callable:
    """
    Applies the equinox filter_grad function to the input params. The 
    corresponding equinox docs are found [here](https://docs.kidger.site/
    equinox/api/filtering/transformations/)


    Parameters
    ----------
    params : Union[PyTree, list, str]
        The params to filter. Can either be a single string path, a list of 
        paths, or a pytree with binary leaves denoting which argument to take
        gradients with respect to.
    *filter_args : Any
        The args to pass to the equinox filter_grad function.
    **filter_kwargs : Any
        The kwargs to pass to the equinox filter_grad function.
    
    Returns
    -------
    Callable
        The wrapped function.
    """
    def wrapper(func : Callable):
        
        @wraps(func)
        def inner_wrapper(pytree : PyTree, *args, **kwargs):

            # Convert params to filter spec if not already
            filter_spec = _convert_to_filter(pytree, params)

            # Wrap original function
            @equinox.filter_grad(*filter_args, **filter_kwargs)
            def recombine(diff : PyTree, non_diff : PyTree):
                return func(combine(diff, non_diff), *args, **kwargs)
            
            # Return wrapped function
            return recombine(*partition(pytree, filter_spec))
        return inner_wrapper
    return wrapper


def filter_value_and_grad(params : Union[PyTree, list, str], 
                *filter_args, **filter_kwargs) -> Callable:
    """
    Applies the equinox filter_value_and_grad function to the input params. The 
    corresponding equinox docs are found [here](https://docs.kidger.site/
    equinox/api/filtering/transformations/)

    Parameters
    ----------
    params : Union[PyTree, list, str]
        The params to filter. Can either be a single string path, a list of 
        paths, or a pytree with binary leaves denoting which argument to take
        gradients with respect to.
    *filter_args : Any
        The args to pass to the equinox filter_value_and_grad function.
    **filter_kwargs : Any
        The kwargs to pass to the equinox filter_value_and_grad function.
    
    Returns
    -------
    Callable
        The wrapped function.
    """
    def wrapper(func : Callable):

        @wraps(func)
        def inner_wrapper(pytree : PyTree, *args, **kwargs):

            # Convert params to filter spec if not already
            filter_spec = _convert_to_filter(pytree, params)

            # Wrap original function
            @equinox.filter_value_and_grad(*filter_args, **filter_kwargs)
            def recombine(diff : PyTree, non_diff : PyTree):
                return func(combine(diff, non_diff), *args, **kwargs)
            
            # Return wrapped function
            return recombine(*partition(pytree, filter_spec))
        return inner_wrapper
    return wrapper


# Dictionary of replaced functions
replaced_dict = {
    'filter_grad': filter_grad,
    'filter_value_and_grad': filter_value_and_grad
}


# Use the __all__ attribute of the external package to get a list of all 
# public functions
external_functions = [func_name for func_name in dir(equinox) 
    if not func_name.startswith("_")]


# Define what function we want to overwrite
replace = ['filter_grad', 'filter_value_and_grad']


# Create a dictionary of wrapper functions that simply call the corresponding 
# function from the external package
wrapper_functions = {}
replaced_functions = []
for func_name in external_functions:
    param = getattr(equinox, func_name)

    # Import functions in the namespace and push them to zodiax namespace
    if callable(param):
        replaced_functions.append(func_name)
        if func_name in replaced_dict.keys():
            wrapper_functions[func_name] = replaced_dict[func_name]
        else:
            wrapper_functions[func_name] = getattr(equinox, func_name)
    
    # If it is a module import it as a submodule
    elif isinstance(param, ModuleType):
        wrapper_functions[func_name] = getattr(equinox, func_name)
    
    # The rest of these should just be custom types that we dont need
    else:
        pass


# Export the wrapper functions from the module
__all__ = replaced_functions
globals().update(wrapper_functions)


import zodiax
from jax.tree_util import tree_map
from equinox import is_array, filter as eqx_filter
from optax import adam, multi_transform, GradientTransformation
from typing import Union
from jaxtyping import PyTree


__all__ = ["get_optimiser"]


def get_optimiser(pytree     : PyTree,
                  paths      : Union[str, list],
                  optimisers : Union[GradientTransformation, list],
                  ) -> tuple:
    """
    Returns an Optax.GradientTransformion object, with the optimisers
    specified by optimisers applied to the leaves specified by paths.

    Parameters
    ----------
    paths : Union[str, list]
        A path or list of paths or list of nested paths.
    optimisers : Union[optax.GradientTransformation, list]
        A optax.GradientTransformation or list of
        optax.GradientTransformation objects to be applied to the leaves
        specified by paths.

    Returns
    -------
    (optimiser, state) : tuple
        A tuple of (Optax.GradientTransformion, optax.MultiTransformState)
        objects, with the optimisers applied to the leaves specified by
        paths, and the initialised optimisation state.
    """
    # Pre-wrap single inputs into a list since optimisers have a length of 2
    if not isinstance(optimisers, list):
        optimisers = [optimisers]
    
    # Paths have to default be wrapped in a list to match optimiser
    if isinstance(paths, str):
        paths = [paths]

    # Construct groups and get param_spec
    groups = [str(i) for i in range(len(optimisers))]
    param_spec = tree_map(lambda _: "null", pytree)
    param_spec = param_spec.set(paths, groups)

    # Generate optimiser dictionary
    opt_dict = dict([(groups[i], optimisers[i]) \
                        for i in range(len(groups))])

    # Assign the null group
    # TODO: Can this be set to None?
    opt_dict["null"] = adam(0.0)

    # Get optimiser object
    optim = multi_transform(opt_dict, param_spec)

    # Get filtered optimiser
    opt_state = optim.init(eqx_filter(pytree, is_array))

    # Return
    return (optim, opt_state)


import zodiax
import jax.numpy as np
from jax.tree_util import tree_map
from equinox import partition, combine
from typing import Union, List


__all__ = ['get_jit_model']


def _float_from_0d(leaf):
    """
    Turns a 0d array into a float for hashing under jit.
    """
    if isinstance(leaf, np.ndarray):
        return float(leaf) if leaf.ndim == 0 else leaf
    else:
        return leaf


def get_jit_model(model : zodiax.base.Base, params : Union[List[str], str]):
    """
    Replaces all scalar array leaves with python floats. This allows for those
    leaves to be marked as static under jit. The arguments in the params input
    are not cast to python types to prevent the model from being recompiled
    every time the arguments change under an optimisation loop.

    Parameters
    ----------
    model : zodiax.Base
        The model to be jitted.
    params : Union[List[str], str]
        The arguments to be optimised under the jit compilation.
    
    Returns
    -------
    model : zodiax.Base
        The model with scalar arrays leaves replaced with python floats.
    """
    args = zodiax.optimisation._convert_to_filter(model, params)
    opt, non_opt = partition(model, args)
    float_model = tree_map(_float_from_0d, non_opt)
    return combine(opt, float_model)


name = "experimental"

# Import as modules
from . import serialisation
from . import jit

# Import core functions from modules
from .serialisation import *
from .jit import *

# Add to __all__
__all__ = serialisation.__all__ + jit.__all__


import setuptools
import os
import codecs
import re

long_description = "Extending Equinox for scientific programming"

here = os.path.abspath(os.path.dirname(__file__))
def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

# DEPENDENCIES
# 1. What are the required dependencies?
with open('requirements.txt') as f:
    install_requires = f.read().splitlines()
# 2. What dependencies required to run the unit tests? (i.e. `pytest --remote-data`)
# tests_require = ['pytest', 'pytest-cov', 'pytest-remotedata']


setuptools.setup(
    python_requires='>=3.8, <4',
    name="zodiax",
    version=find_version("zodiax", "__init__.py"),
    description="Equinox extension for scientific programming",
    long_description=long_description,

    author="Louis Desdoigts",
    author_email="Louis.Desdoigts@sydney.edu.au",
    url="https://github.com/LouisDesdoigts/zodiax",

    project_urls={
        "Bug Tracker": "https://github.com/LouisDesdoigts/zodiax/issues",
    },

    install_requires=install_requires,

    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],

    packages = ["zodiax", "zodiax/experimental"],
)



from __future__ import annotations
import sys
import pickle
import zodiax
from typing import Any
from jax import Array
# from jax.typing import ArrayLike
import numpy as onp
import jax.numpy as jnp
from equinox import tree_serialise_leaves, tree_deserialise_leaves
from pathlib import Path


__all__ = ['serialise', 'deserialise', 'load_structure', 'build_structure']


#####################
### Serialisation ###
#####################
def _check_node(obj      : Any, 
                self_key : str  = None, 
                depth    : int  = 0., 
                _print   : bool = False) -> bool:
    """
    Checks if the input object is a container or a leaf node. If the object is
    a leaf False is returned else True.

    Container types:
        list, tuple, dict, zodiax.Base
    
    Leaf types:
        jax.ArrayLike, numpy.ndarray, bool, complex, float, int, str, None
    
    All other types will raise an error as they are not supported, but likely
    could be supported in future.

    Parameters
    ----------
    obj : Any
        The object to check.
    depth : int = 0
        The depth of the object in the tree. Use to print the tree structure
        for debugging.
    self_key : str = None
        The key of the object in the parent container. Use to print the tree
        structure for debugging.
    _print : bool = False
        If True, print the tree structure for debugging.
    
    Returns
    -------
    is_container : bool
        True if the object is a container, False if it is a leaf.
    """
    t = '  ' * depth
    conatiner_types = (list, tuple, dict, zodiax.Base)
    leaf_types = (Array, onp.ndarray, onp.bool_, onp.number, bool, int, float,
        complex, str)

    # Contianer node
    if isinstance(obj, conatiner_types):
        if _print: print(f"{t}Node '{self_key}' of type: {type(obj)}")
        return True

    # NOTE: Checking None types in a tuple of types has some off behaviour. 
    # There have been changes from python 3.7 - 3.8 and then 3.9-3.10.
    # This seems to be the most robust way to check for None types.

    # Leaf node
    elif isinstance(obj, leaf_types) or isinstance(obj, type(None)):
        if _print: print(f"{t}Leaf '{self_key}' of type: {type(obj)}")
        return False

    # Unknown - Raise error
    else:
        raise ValueError(f"Unknown node '{self_key}' of type: {type(obj)}. "
            "Please raise an issue on GitHub to enable support for this type.")
    

def _get_accessor(obj : Any) -> tuple:
    """
    Returns the keys and accessor for the input object.

    Parameters
    ----------
    obj : Any
        The object to get the keys and accessor for.

    Returns
    -------
    (keys, accessor) : tuple
        keys : list
            The keys of the object.
        accessor : function
            The accessor function for the object.
    """
    accessor = lambda object, x: object[x]

    # Lists and tuples
    if isinstance(obj, (list, tuple)):
        keys = list(range(len(obj)))
        return keys, accessor

    # Zodaix object types
    elif isinstance(obj, zodiax.Base):
        accessor = lambda object, x: getattr(object, x)
        keys = obj.__dataclass_fields__.keys()
        return keys, accessor

    # Dictionary types
    else:
        return obj.keys(), accessor


def _format_type(obj : Any) -> str:
    """
    Returns a string of the object type with the extra '< class ' and '>' 
    removed.

    Parameters
    ----------
    obj : Any
        The object to get the type of.
    
    Returns
    -------
    type_str : str
        The string of the object type.
    """
    class_str = str(type(obj))
    return class_str[class_str.find("'")+1:class_str.rfind("'")]


def _format_dtype(dtype : onp.dtype) -> str:
    """
    Formats the array dtype into a string. Supports boolean, integer, float 
    and complex types.

    Parameters
    ----------
    dtype : numpy.dtype
        The array dtype.
    
    Returns
    -------
    dtype_str : str
    """
    kind_map = {'b': 'bool', 'f': 'float', 'i': 'int', 'c': 'complex'}
    return f"{kind_map[dtype.kind]}{8*dtype.itemsize}"


def _build_node(obj    : Any, 
               key    : str  = None, 
               depth  : int  = 0, 
               _print : bool = False) -> dict:
    """
    Builds a node for the input object.

    Parameters
    ----------
    obj : Any
        The object to get the leaves of.
    key : str
        The key of the object in the parent container. Use to print the tree
        structure for debugging.
    depth : int
        The depth of the object in the tree. Use to print the tree structure
        for debugging.
    _print : bool
        If True, print the tree structure for debugging.
    
    Returns
    -------
    node_dict : dict
        The dictionary detailing the type and metadata of the node.
    """
    inner_structure = build_structure(obj, key, depth+1, _print)

    # Container node
    if isinstance(inner_structure, dict):
        return _build_conatiner_node(obj, inner_structure)

    # Leaf node
    else:
        return _build_leaf_node(obj)
    

def _build_leaf_node(obj : Any) -> dict:
    """
    Builds a leaf node for the input object. Explicity handles string and array
    types. Strings are stored directly and serialised in the structure 
    dictionary and arrays have their shape and dtypes serialised for 
    reconstruction using the `equinox.deserialise_tree_leaves`.

    Parameters
    ----------
    obj : Any
        The object to get the leaves of.
    
    Returns
    -------
    node_dict : dict
        The dictionary detailing the type and metadata of the leaf.
    """
    # Basic info
    node_dict = {'node_type': 'leaf', 
                 'type': _format_type(obj)}
    
    # Handle string case
    if isinstance(obj, str):
        node_dict['value'] = obj
    
    # Append meta-data for arrays
    elif isinstance(obj, (Array, onp.ndarray)):
        node_dict['shape'] = obj.shape
        node_dict['dtype'] = _format_dtype(obj.dtype)
    
    return node_dict


def _build_conatiner_node(obj : Any, inner_structure : dict) -> dict:
    """
    Builds a container node for the input object.

    Parameters
    ----------
    obj : Any
        The object to build the container node for.
    inner_structure : dict
        The dictionary detailing the structure of the object.

    Returns
    -------
    node_dict : dict
        The dictionary detailing the type and metadata of the container.
    """
    return {'node_type': 'container', 
            'type': _format_type(obj), 
            'node': inner_structure}


def build_structure(obj      : Any, 
                    self_key : str  = None, 
                    depth    : int  = 0,
                    _print   : bool = False):
    """
    Recursively iterates over the input object in order to return a dictionary 
    detailing the strucutre of the of the object. Each node can be either a
    conainter node or leaf node. Each node is a dictionary with the following
    structure:

    {'node_type': 'container' or 'leaf',
     'type': str,
     'node': {
        param1 : {'node_type' : 'container', ...}, -> If container
        param2 : {'node_type' : 'leaf',
                  '...' : ...}, -> If leaf conatining any leaf metadata
        }
    
    Specific leaf metadata:
        Strings:
            String values are stored in the 'value' key and serialised via the
            returned structure dictionary.
        Jax/Numpy Arrays:
            Both the array shape and dtype are stored in the 'shape' and
            'dtype' keys respectively. 
    
    This method can be developed further to support more leaf types, since each
    individual leaf type can be made to store any arbitrarity metadata, as long
    as it can be serialised by json and used to deserialise it later.
    
    This dictionary can then be serialised using pickle and then later
    used to deserialise the object in conjunction with equinox leaf 
    serialise/deserialise methods.

    NOTE: This method is not equipped to handle `equinox.static_field()` 
    parameters, as they can be arbitrary data types but do not get serialised
    by the  `equinox.serialise_tree_leaves()` methods and hence require custom 
    serialisation via this method. Therefore this method currently does not
    handle this case correctly. This is not checked for currently so will
    silently break or result in unexpected behaviour.

    TODO: Serialise package versions in order to raise warnings when 
    deserialising about inconsistent versions.

    Parameters
    ----------
    obj : Any
        The object to get the leaves of.
    self_key : str = None
        The key of the object in the parent container. Use to print the tree
        structure for debugging.
    depth : int = 0
        The depth of the object in the tree. Use to print the tree structure
        for debugging.
    _print : bool = False
        If True, print the tree structure for debugging.
    
    Returns
    -------
    structure : dict
        The dictionary detailing the structure of the object.
    """
    structure = {}
    is_container = _check_node(obj, self_key, depth, _print=_print)

    # Recursive case
    if is_container:
        keys, accessor = _get_accessor(obj)

        # Iterate over parameters
        for key in keys:
            sub_obj = accessor(obj, key)
            structure[key] = _build_node(sub_obj, key, depth, _print)
                    
        # Deal with outermost container
        if depth == 0:
            return _build_conatiner_node(obj, structure)
        else:
            return structure

    # Base case    
    else:
        return obj


def serialise(path : str, obj : Any) -> None:
    """
    Serialises the input zodiax pytree to the input path. This method works by
    creating a dictionary detailing the structure of the object to be
    serialised. This dictionary is then serialised using `pickle` and the 
    pytree leaves are serialised using `equinox.serialise_tree_leaves()`. This
    object can then be deserialised using the `deserialise()` method. 

    This method is currently considered experimental for a number of reasons:
     - Some objects can not be gaurenteed to be deserialised correctly. 
     - User-defined classes _can_ be serialised but it is up to the user to 
     import the class into the global namespace when deserialising. 
     - User defined functions can not be gaurenteed to be deserialised
     correctly.
     - Different versions of packages can cause issues when deserialising. This
    metadata is planned to be serialised in the future and have warnings raised
    when deserialising.
     - static_field() parameters are not handled correctly. Since array types
     can be set as static_field() parameters, they are not serialised by
     `equinox.serialise_tree_leaves()` and hence require custom serialisation
     via this method. This is not checked for currently so will silently break.
     This can be fixed with some pre-filtering and type checking using the
     `.tree_flatten()` method.

    Parameters
    ----------
    path : str
        The path to serialise the object to.
    obj : Any
        The object to serialise.
    """
    # Check path type
    if not isinstance(path, (str, Path)):
        raise TypeError(f'path must be a string or Path, not {type(path)}')
    else:
        # Convert to string in case of Path for adding .zdx extension
        path = str(path)
    
    # Add default .zdx extension
    if len(path.split('.')) == 1:
        path += '.zdx'
    
    # Serialise
    structure = build_structure(obj)
    with open(path, 'wb') as f:
        pickle.dump(structure, f)
        tree_serialise_leaves(f, obj)
    

#######################
### Deserialisation ###
#######################
def _construct_class(modules_str : str) -> object:
    """
    Constructs an empty instance of some class from a string of the form
    'module.sub_module.class'. 

    Explicitly handled types: Nones, bool, int, float, str, complex, list, 
    tuple, dict, jax arrays, numpy arrays.

    Other types are imported and attempted to be instantiated.


    Parameters
    ----------
    modules_str : str
        The string of the form 'module.sub_module.class' to construct

    Returns
    -------
    object : object
        The instantiated class
    """
    # None case
    if modules_str == 'NoneType':
        return None

    # Regular python types
    elif modules_str in ['bool', 'int', 'float', 'str', 'complex', 'list', 
                         'tuple', 'dict']:
        return eval(f"{modules_str}()")

    # Array types
    elif modules_str == 'jaxlib.xla_extension.Array':
        import jax.numpy as jnp
        return jnp.array([])
    elif modules_str == 'numpy.ndarray':
        import numpy as onp
        return onp.array([])

    # Modules that existed in the global namespace - Search for it
    elif modules_str.startswith('__main__'):
        modules = modules_str.split('.')[1:]
        try:
            module = getattr(sys.modules['__main__'], modules[0])
            if len(modules) == 1:
                return module.__new__(module)
            else:
                for sub_module in modules:
                    module = getattr(module, sub_module)
                return module.__new__(module)
        except AttributeError:
            raise AttributeError(f"The module/class '{'.'.join(modules)}' "
                "originally existed in the global namespace, but does not "
                "exist in this script. It must be imported in order to be "
                "initialised.")
    
    # Finally attempt import of module
    else:
        from importlib import import_module as im
        modules = modules_str.split('.')

        if len(modules) == 1:
            raise ValueError(f"Unhandled case: {modules}")
        elif len(modules) == 2:
            cls = getattr(im(f'{modules[0]}'), modules[-1])
        else:
            sub_modules = '.'.join(modules[1:-1])
            cls = getattr(im(f'{modules[0]}.{sub_modules}'), modules[-1])
        return cls.__new__(cls)


def _load_container(obj : object, key : str, value : object) -> object:
    """
    Updates and return the object with the supplied key and value pair.

    TODO: Check the setting of different types of containers, ie list
     -> What about tuples which are immutable?

    Parameters
    ----------
    obj : object
        The object to update.
    key : str
        The key at which to update.
    value : object
        The corresponding value to update with.
    
    Returns
    -------
    obj : object
        The updated object.
    """
    # Recursively load the sub node
    sub_node = load_structure(value)

    # Object case
    if isinstance(obj, zodiax.Base):
        object.__setattr__(obj, key, sub_node)
    
    # List/tuple case
    # TODO: Check this preserves order
    # TODO: Check original input type and re-cast to tuple if necessary
    elif isinstance(obj, list):
        obj.append(sub_node)

    # Dict case
    else:
        obj[key] = sub_node
    
    return obj


def _load_leaf(obj : object, structure : dict) -> object:
    """
    Returns the leaf value, handing the special cases of strings and arrays.
    Can be expanded to deal with mode complex cases.

    Parameters
    ----------
    obj : object
        The object to update.
    structure : dict
        The structure to load.

    Returns
    -------
    obj : object
        The updated object.
    """
    # String case
    if isinstance(obj, str):
        obj = structure['value']

    # Jax Array casees
    elif isinstance(obj, Array):
        dtype = getattr(jnp, structure['dtype'])
        obj = jnp.zeros(structure['shape'], dtype=dtype)

    # Numpy Array case
    elif isinstance(obj, onp.ndarray):
        dtype = getattr(onp, structure['dtype'])
        obj = onp.zeros(structure['shape'], dtype=dtype)
    
    # All others
    else:
        pass
    
    return obj


def load_structure(structure : dict) -> object:
    """
    Load a structure from a dictionary to later be used in conjuction with
    `eqx.tree_deserialise_leaves()`.


    Custom leaf node desrialisation is handled by the `_load_leaf` function.
    
    Parameters
    ----------
    structure : dict
        The structure to load.

    Returns
    -------
    obj : object
        The loaded structure.
    """
    # Construct the object
    obj = _construct_class(structure['type'])

    # Container Node
    if structure['node_type'] == 'container': 
        
        # Iterarte over all parameters and update the object
        for key, value in structure['node'].items():
            obj = _load_container(obj, key, value)
        return obj
    
    # Leaf Node
    else: 
        return _load_leaf(obj, structure)


def deserialise(path : str):
    """
    Deserialises the input object at the input path.

    Parameters
    ----------
    path : str
        The path to serialise the object to.
    
    Returns
    -------
    obj : Any
        The deserialised object.
    """
    with open(path, 'rb') as f:
        structure = pickle.load(f)
        like = load_structure(structure)
        obj = tree_deserialise_leaves(f, like)
    return obj


from jax.tree_util import tree_map
from typing import Union
from jaxtyping import PyTree


__all__ = ["get_args"] #, "get_param_spec"]


def get_args(pytree: PyTree, paths : Union[str, list]) -> PyTree:
    """
    Returns a pytree of matching structure with boolean values at the leaves.
    Leaves specified by paths will be True, all others will be False. 

    Parameters
    ----------
    paths : Union[str, list]
        A path or list of paths or list of nested paths.

    Returns
    -------
    args : PyTree
        An pytree of matching structre with boolean values at the leaves.
    """
    args = tree_map(lambda _: False, pytree)
    paths = paths if isinstance(paths, list) else [paths]
    values = len(paths) * [True]
    return args.set(paths, values)


# def get_param_spec(pytree   : PyTree,
#                    paths    : Union[str, list],
#                    groups   : Union[str, list],
#                    get_args : bool = False) -> PyTree:
#     """
#     Returns 'param_spec' object, to be used in conjunction with the
#     Optax.multi_transform functions. The param_spec is a pytree of matching
#     strucutre that has strings assigned to every node, denoting the group
#     that is belongs to. Each of these groups can then have unique optimiser
#     objects assigned to them. This is typically used to assign different
#     learning rates to different parameters.

#     Note this sets the default or non-trainable group to 'null'.

#     Parameters
#     ----------
#     paths : Union[str, list]
#         A path or list of paths or list of nested paths.
#     groups : Union[str, list]
#         A string or list of strings, denoting which group to assign the
#         corresponding leaves denoted by paths to.
#     get_args : bool = False
#         Return a corresponding args pytree or not.

#     Returns
#     -------
#     param_spec : PyTree
#         An pytree of matching structre with string values at the leaves
#         specified by groups.
#     """


#     return param_spec if not get_args \
#         else (param_spec, pytree.get_args(paths))